defmodule Archeometer.Analysis.Apps.Xref do
  @moduledoc """
  Functions for generating a dependency graph from applications in an Elixir
  project.

  Accepted output formats are "dot" (graphviz), "png" and "mermaid".
  """

  alias Archeometer.Graphs.{Graphviz, Mermaid}
  alias Archeometer.Explore.Project

  @supported_formats ["png", "dot", "mermaid", "svg"]

  @doc """
  Creates a dependency graph between the applications of the current project.

  ## Parameters

  - `format` can be one of "dot" (graphviz), "png", or "mermaid".

  ## Returns

  - The binary representing the graph, if the operation was completed successfully.
  - `{:error, reason}` if not.

  """
  def gen_graph(format, extra_deps \\ [])

  def gen_graph(format, extra_deps) when format in @supported_formats do
    do_gen_graph(format, Project.apps() ++ extra_deps)
  end

  def gen_graph(_format, _apps) do
    {:error, :unsupported_format}
  end

  defp do_gen_graph(format, apps) do
    apps
    |> Enum.flat_map(fn app ->
      case filtered_deps(app, apps) do
        {:error, _} -> []
        {:ok, deps} -> [{app, deps}]
      end
    end)
    |> Enum.into(%{})
    |> render(format)
  end

  defp filtered_deps(app, full_apps) do
    with {:ok, deps} <- Project.get_app_deps(app) do
      {:ok, Enum.filter(deps, &(&1 in full_apps))}
    end
  end

  defp render(refs, "dot") do
    Graphviz.render_dot(refs)
  end

  defp render(refs, format) when format in ["png", "svg"] do
    Graphviz.render_image(refs, format)
  end

  defp render(refs, "mermaid") do
    Mermaid.render(refs)
  end
end
